
###############################################################################
##                                                                           ##
##    Copyright (c) 1995 - 2004 by Steffen Beyer.                            ##
##    All rights reserved.                                                   ##
##                                                                           ##
##    This package is free software; you can redistribute it                 ##
##    and/or modify it under the same terms as Perl itself.                  ##
##                                                                           ##
###############################################################################

package Date::Pcalc;

BEGIN { eval { require bytes; }; }
use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);
use constant YEAR_OF_EPOCH => 70;
use constant CENTURY_OF_EPOCH => 1900;
use constant EPOCH => (CENTURY_OF_EPOCH + YEAR_OF_EPOCH);
use constant LANGUAGES => 14;

require Exporter;
require DynaLoader;

@ISA = qw(Exporter DynaLoader);

@EXPORT = qw();

@EXPORT_OK = qw(
    Days_in_Year
    Days_in_Month
    Weeks_in_Year
    leap_year
    check_date
    check_time
    check_business_date
    Day_of_Year
    Date_to_Days
    Day_of_Week
    Week_Number
    Week_of_Year
    Monday_of_Week
    Nth_Weekday_of_Month_Year
    Standard_to_Business
    Business_to_Standard
    Delta_Days
    Delta_DHMS
    Delta_YMD
    Delta_YMDHMS
    Normalize_DHMS
    Add_Delta_Days
    Add_Delta_DHMS
    Add_Delta_YM
    Add_Delta_YMD
    Add_Delta_YMDHMS
    System_Clock
    Today
    Now
    Today_and_Now
    This_Year
    Gmtime
    Localtime
    Mktime
    Timezone
    Date_to_Time
    Time_to_Date
    Easter_Sunday
    Decode_Month
    Decode_Day_of_Week
    Decode_Language
    Decode_Date_EU
    Decode_Date_US
    Fixed_Window
    Moving_Window
    Compress
    Uncompress
    check_compressed
    Compressed_to_Text
    Date_to_Text
    Date_to_Text_Long
    English_Ordinal
    Calendar
    Month_to_Text
    Day_of_Week_to_Text
    Day_of_Week_Abbreviation
    Language_to_Text
    Language
    Languages
    Decode_Date_EU2
    Decode_Date_US2
    Parse_Date
    ISO_LC
    ISO_UC
);

%EXPORT_TAGS = (all => [@EXPORT_OK]);

##################################################
##                                              ##
##  "Version()" is available but not exported   ##
##  in order to avoid possible name clashes.    ##
##  Call with "Date::Calc::Version()" instead!  ##
##                                              ##
##################################################

$VERSION = '5.4';

bootstrap Date::Calc $VERSION;

@days_in_year = (
	[ 0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 ],
    [ 0, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 ]
);

@days_in_month = (
	[ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ],
    [ 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]
);

$language = 1;	# default = 1 (English)

@month_to_text = (
    [
        "???", "???", "???", "???", "???", "???", "???",
        "???", "???", "???", "???", "???", "???"
    ],
    [
        "???", "January", "February", "March", "April", "May", "June",
        "July", "August", "September", "October", "November", "December"
    ],
    [
        "???", "janvier", "f�vrier", "mars", "avril", "mai", "juin",
        "juillet", "ao�t", "septembre", "octobre", "novembre", "d�cembre"
    ],
    [
        "???", "Januar", "Februar", "M�rz", "April", "Mai", "Juni",
        "Juli", "August", "September", "Oktober", "November", "Dezember"
    ],
    [
        "???", "enero", "febrero", "marzo", "abril", "mayo", "junio",
        "julio", "agosto", "septiembre", "octubre", "noviembre", "diciembre"
    ],
    [
        "???", "janeiro", "fevereiro", "mar�o", "abril", "maio", "junho",
        "julho", "agosto", "setembro", "outubro", "novembro", "dezembro"
    ],
    [
        "???", "januari", "februari", "maart", "april", "mei", "juni",
        "juli", "augustus", "september", "october", "november", "december"
    ],
    [
        "???", "Gennaio", "Febbraio", "Marzo", "Aprile", "Maggio", "Giugno",
        "Luglio", "Agosto", "Settembre", "Ottobre", "Novembre", "Dicembre"
    ],
    [
        "???", "januar", "februar", "mars", "april", "mai", "juni",
        "juli", "august", "september", "oktober", "november", "desember"
    ],
    [
        "???", "januari", "februari", "mars", "april", "maj", "juni",
        "juli", "augusti", "september", "oktober", "november", "december"
    ],
    [
        "???", "januar", "februar", "marts", "april", "maj", "juni",
        "juli", "august", "september", "oktober", "november", "december"
    ],
    [
        "???", "tammikuu", "helmikuu", "maaliskuu", "huhtikuu",
        "toukokuu", "kes�kuu", "hein�kuu", "elokuu",
        "syyskuu", "lokakuu", "marraskuu", "joulukuu"
    ],
    [
        "???", "Janu�r", "Febru�r", "M�rcius", "�prilis", "M�jus", "J�nius",
        "J�lius", "Augusztus", "Szeptember", "Okt�ber", "November", "December"
    ],
    [
        "???", "stycznia", "lutego", "marca", "kwietnia", "maja", "czerwca",
        "lipca", "sierpnia", "wrzesnia", "pazdziernika", "listopada", "grudnia"
     # "???", "Styczen", "Luty", "Marzec", "Kwiecien", "Maj", "Czerwiec", # non-flected?
     # "Lipiec", "Sierpien", "Wrzesien", "Pazdziernik", "Listopad", "Grudzien"
     # "???", "Stycze�", "Luty", "Marzec", "Kwiecie�", "Maj", "Czerwiec", # ISO-Latin-2
     # "Lipiec", "Sierpie�", "Wrzesie�", "Pa�dziernik", "Listopad", "Grudzie�" 
    ],
    [
        "???", "Ianuarie", "Februarie", "Martie", "Aprilie", "Mai", "Iunie",
        "Iulie", "August", "Septembrie", "Octombrie", "Noiembrie", "Decembrie"
    ]
);

@day_of_week_to_text = (
    [
        "???", "???", "???", "???",
        "???", "???", "???", "???"
    ],
    [
        "???", "Monday", "Tuesday", "Wednesday",
        "Thursday", "Friday", "Saturday", "Sunday"
    ],
    [
        "???", "Lundi", "Mardi", "Mercredi",
        "Jeudi", "Vendredi", "Samedi", "Dimanche"
    ],
    [
        "???", "Montag", "Dienstag", "Mittwoch",
        "Donnerstag", "Freitag", "Samstag", "Sonntag"
    ],
    [
        "???", "Lunes", "Martes", "Mi�rcoles",
        "Jueves", "Viernes", "S�bado", "Domingo"
    ],
    [
        "???", "Segunda-feira", "Ter�a-feira", "Quarta-feira",
        "Quinta-feira", "Sexta-feira", "S�bado", "Domingo"
    ],
    [
        "???", "Maandag", "Dinsdag", "Woensdag",
        "Donderdag", "Vrijdag", "Zaterdag", "Zondag"
    ],
    [
        "???", "Luned�", "Marted�", "Mercoled�",
        "Gioved�", "Venerd�", "Sabato", "Domenica"
    ],
    [
        "???", "mandag", "tirsdag", "onsdag",
        "torsdag", "fredag", "l�rdag", "s�ndag"
    ],
    [
        "???", "m�ndag", "tisdag", "onsdag",
        "torsdag", "fredag", "l�rdag", "s�ndag"
    ],
    [
        "???", "mandag", "tirsdag", "onsdag",
        "torsdag", "fredag", "l�rdag", "s�ndag"
    ],
    [
        "???", "maanantai", "tiistai", "keskiviikko",
        "torstai", "perjantai", "lauantai", "sunnuntai"
    ],
    [
        "???", "h�tf�", "kedd", "szerda",
        "cs�t�rt�k", "p�ntek", "szombat", "vas�rnap"
    ],
    [
        "???", "poniedzialek", "wtorek", "srodek",
        "czwartek", "piatek", "sobota", "niedziele"
     #  "???", "Poniedzialek", "Wtorek", "Sroda", # non-flected
     #  "Czwartek", "Piatek", "Sobota", "Niedziela"
     #  "???", "Poniedzia�ek", "Wtorek", "�roda", # ISO-Latin-2
     # "Czwartek", "Pi�tek", "Sobota", "Niedziela"
    ],
    [
        "???", "Luni", "Marti", "Miercuri",
        "Joi", "Vineri", "Sambata", "Duminica"
    ]
);

@day_of_week_abbreviation = (
    [
        "", "", "", "", "", "", "", ""    #   0 
    ],
    [
        "", "", "", "", "", "", "", ""    #   1 
    ],
    [
        "", "", "", "", "", "", "", ""    #   2 
    ],
    [
        "", "", "", "", "", "", "", ""    #   3 
    ],
    [
        "", "", "", "", "", "", "", ""    #   4 
    ],
    [
        "???", "2�", "3�", "4�", "5�", "6�", "S�b", "Dom"    #   5 
    ],
    [
        "", "", "", "", "", "", "", ""    #   6 
    ],
    [
        "", "", "", "", "", "", "", ""    #   7 
    ],
    [
        "", "", "", "", "", "", "", ""    #   8 
    ],
    [
        "", "", "", "", "", "", "", ""    #   9 
    ],
    [
        "", "", "", "", "", "", "", ""    #  10 
    ],
    [
        "", "", "", "", "", "", "", ""    #  11 
    ],
    [
        "", "", "", "", "", "", "", ""    #  12 
    ],
    [
        "???", "Pn", "Wt", "Sr", "Cz", "Pt", "So", "Ni"    #  13 
     #  "???", "Pn", "Wt", "�r", "Cz", "Pt", "So", "Ni" # 13 ISO-Latin-2 
    ],
    [
        "", "", "", "", "", "", "", ""    #  14 
    ]

);

@english_ordinals = qw( th st nd rd );

@date_long_format = (
    "%s, %d %s %d",                     #   0  Default     
    "%s, %s %s %d",                     #   1  English     
    "%s %d %s %d",                      #   2  Fran�ais    
    "%s, den %d. %s %d",                #   3  Deutsch     
    "%s, %d de %s de %d",               #   4  Espa�ol     
    "%s, dia %d de %s de %d",           #   5  Portugu�s   
    "%s, %d %s %d",                     #   6  Nederlands  
    "%s, %d %s %d",                     #   7  Italiano    
    "%s, %d. %s %d",                    #   8  Norsk       
    "%s, %d %s %d",                     #   9  Svenska     
    "%s, %d. %s %d",                    #  10  Dansk       
    "%s, %d. %sta %d",                  #  11  suomi       
    "%d. %s %d., %s",                   #  12  Magyar      
    "%s, %d %s %d",                     #  13  polski      
    "%s %d %s %d"                       #  14  Romaneste   
);

@language_to_text = (
	"???", "English", "Fran�ais", "Deutsch", "Espa�ol",
    "Portugu�s", "Nederlands", "Italiano", "Norsk", "Svenska",
    "Dansk", "suomi", "Magyar", "polski", "Romaneste"
);

use constant DATE_ERROR       => "not a valid date";
use constant TIME_ERROR       => "not a valid time";
use constant YEAR_ERROR       => "year out of range";
use constant MONTH_ERROR      => "month out of range";
use constant WEEK_ERROR       => "week out of range";
use constant DAYOFWEEK_ERROR  => "day of week out of range";
use constant DATE_RANGE_ERROR => "date out of range";
use constant TIME_RANGE_ERROR => "time out of range";
use constant FACTOR_ERROR     => "factor out of range";
use constant LANGUAGE_ERROR   => "language not available";
use constant SYSTEM_ERROR     => "not available on this system";
use constant MEMORY_ERROR     => "unable to allocate memory";
use constant STRING_ERROR     => "argument is not a string";

sub error
{
	my $message = shift;
	my $caller = (caller(1))[3]
	croak( "$caller(): $message");
}

sub center
{
	my ($target_ref, $source, $width) = @_;
	my $length = length $source;
	my $n_blanks;

	$length = $width if ($length > $width);
	$n_blanks = ($width - $length) >> 1;
	${$target_ref} = " " x $n_blanks . $source . "\n";
}

sub blank
{
	my ($target_ref, $count) = @_;
	${$target_ref} .= " " x $count;
}

sub newline
{
	my ($target_ref, $count) = @_;
	${$target_ref} .= "\n" x $count;
}

sub normalize_time
{
	my ($dd, $dh, $dm, $ds) = @_;
	my $quot;
	$quot = int( ${$ds} / 60 );
	${$ds} %= 60;
	${$dm} += $quot;
	
	$quot = int( ${$dm} / 60 );
	${$dm} %= 60;
	${$dh} += $quot;
	
	$quot = int( ${$dh} / 24 );
	${$dh} %= 24;
	${$dd} += $quot;
}

sub normalize_ranges
{
	my ($dd, $dh, $dm, $ds) = @_;
	my $quot;

	$quot = int( ${$dh} / 24 );
	${dh} %= 24;
	${$dd} += $quot;

	$quot = int( ${$dm} / 60 );
	${$dm} %= 60;
	${$dh} += $quot;

	normalize_time( $dd, $dh, $dm, $ds );	
}

sub normalize_signs
{
	my ($dd, $dh, $dm, $ds) = @_;
	my $quot;
	$quot = int( ${$ds} / 86400 );
	${$ds} %= 86400;
	${$dd} += $quot;
	if (${$dd} != 0)
	{
		if (${$dd} > 0)
		{
			if (${$ds} < 0)
			{
				${$ds} += 86400;
				${$dd}--;
			}
		}
		else
		{
			if (${$ds} > 0)
			{
				${$ds} -= 86400;
				${$dd}++;
			}
		}
	}
	${$dh} = 0;
	${$dm} = 0;
	if (${$ds} != 0)
	{
		normalize_time( $dd, $dh, $dm, $ds );
	}
}

sub normalize_DHMS
{
	my ($dd, $dh, $dm, $ds) = @_;
	normalize_ranges( $dd, $dh, $dm, $ds );
	${$ds} += ((${$dh}*60) + ${$dm}) * 60;
	normalize_signs( $dd, $dh, $dm, $ds );
}

sub Days_in_Year
{
	my ($year, $month) = @_;
	if ($year > 0)
	{
		if ($month >= 1 && $month <= 12)
		{
			return $days_in_year[leap_year($year)][$month+1];
		}
		else
		{
			error( MONTH_ERROR );
		}
	}
	else
	{
		error( YEAR_ERROR );
	}
}

sub Days_in_Month
{
	my ($year, $month) = @_;
	if ($year > 0)
	{
		if ($month >= 1 && $month <= 12)
		{
			return $days_in_month[leap_year($year)][$month];
		}
		else
		{
			error( MONTH_ERROR );
		}
	}
	else
	{
		error( YEAR_ERROR );
	}
}

sub Weeks_in_Year
{
	my $year = shift;
	if ($year > 0)
	{
		return (52 + (( Day_of_Week($year, 1, 1) == 4) ||
			(Day_of_week($year, 12, 31) == 4)) );
	}
	else
	{
		error( YEAR_ERROR );
	}
}

sub Week_Number
{
	my ($year, $month, $day) = @_;
	if (check_date($year, $month, $day))
	{
		my $first = Day_of_Week($year, 1, 1) - 1;
		return (
			int(Delta_Days($year, 1, 1, $year, $month, $day) + $first / 7)
			+ (($first < 4) ? 1 : 0)
		);
	}
	else
	{
		error(DATE_ERROR);
	}
}
################################################################
sub leap_year
{
	my $year = shift;
	my $yy;
	if ($year > 0)
	{
		return ( ($year & 0x03 == 0) &&
			( ($year % 100 != 0) || ($year % 400 == 0) ) );
	}
	else
	{
		error( YEAR_ERROR );
	}
}

sub check_date
{
	my ($year, $month, $day) = @_;

	return (($year >= 1) and
        ($month >= 1) and ($month <= 12) and
        ($day >= 1) and
        ($day <= $Days_in_Month[leap_year($year)][$month]));
}

sub check_time
{
	my ($hour, $min, $sec) = @_;
	return ( ($hour >= 0) && ($min >= 0) && ($sec >= 0) &&
		($hour < 24) && ($min < 60) && ($sec < 60) ) ? 1 : 0;
}

sub check_business_date
{
	my ($year, $week, $day_of_week) = @_;
	return (($year >=1 ) && ($week >= 1) &&
		($week <= Weeks_in_Year($year) &&
		($day_of_week >= 1) && ($day_of_week <= 7)) ? 1 : 0;
}

sub Day_of_Year
{
	my ($year, $month, $day) = @_;
	my $result = 0;
	my $leap;
	
	if (($year >= 1) &&
        ($month >= 1) && ($month <= 12) &&
        ($day >= 1) &&
        ($day <= $days_in_month_[$leap=leap_year($year)][$month]))
	{
		$result = $days_in_$ear_[$leap][$month] + $day;
	}
	if ($result == 0)
	{
		error( DATE_ERROR );
	}
	return $result;
}

sub Date_to_Days
{
	my ($year, $month, $day) = @_;
	if (($year >= 1) &&
        ($month >= 1) && ($month <= 12) &&
        ($day >= 1) &&
        ($day <= $days_in_month_[$leap=leap_year($year)][$month]))
	{
		return( Year_to_Days(--$year) +
			$days_in_$ear[$leap][$month] + $day );
	}
	else
	{
		error( DATE_ERROR );
	}
}

sub Day_of_Week
{
	my ($year, $month, $day) = @_;
	my $days = Date_to_Days($year, $month, $day);
    if ($days > 0L)
    {
        $days--;
        $days %= 7;
        $days++;
    }
    if ($days != 0)
	{
		return $days;
	}
	error( DATE_ERROR );
}

sub Decode_Date_EU2
{
    die "Usage: (\$year,\$month,\$day) = Decode_Date_EU2(\$date);\n"
      if (@_ != 1);

    my($buffer) = @_;
    my($year,$month,$day,$length);

    if ($buffer =~ /^\D*  (\d+)  [^A-Za-z0-9\xC0-\xD6\xD8-\xF6\xF8-\xFF]*  ([A-Za-z\xC0-\xD6\xD8-\xF6\xF8-\xFF]+)  [^A-Za-z0-9\xC0-\xD6\xD8-\xF6\xF8-\xFF]*  (\d+)  \D*$/x)
    {
        ($day,$month,$year) = ($1,$2,$3);
        $month = Decode_Month($month);
        unless ($month > 0)
        {
            return(); # can't decode month!
        }
    }
    elsif ($buffer =~ /^\D*  0*(\d+)  \D*$/x)
    {
        $buffer = $1;
        $length = length($buffer);
        if    ($length == 3)
        {
            $day   = substr($buffer,0,1);
            $month = substr($buffer,1,1);
            $year  = substr($buffer,2,1);
        }
        elsif ($length == 4)
        {
            $day   = substr($buffer,0,1);
            $month = substr($buffer,1,1);
            $year  = substr($buffer,2,2);
        }
        elsif ($length == 5)
        {
            $day   = substr($buffer,0,1);
            $month = substr($buffer,1,2);
            $year  = substr($buffer,3,2);
        }
        elsif ($length == 6)
        {
            $day   = substr($buffer,0,2);
            $month = substr($buffer,2,2);
            $year  = substr($buffer,4,2);
        }
        elsif ($length == 7)
        {
            $day   = substr($buffer,0,1);
            $month = substr($buffer,1,2);
            $year  = substr($buffer,3,4);
        }
        elsif ($length == 8)
        {
            $day   = substr($buffer,0,2);
            $month = substr($buffer,2,2);
            $year  = substr($buffer,4,4);
        }
        else { return(); } # wrong number of digits!
    }
    elsif ($buffer =~ /^\D*  (\d+)  \D+  (\d+)  \D+  (\d+)  \D*$/x)
    {
        ($day,$month,$year) = ($1,$2,$3);
    }
    else { return(); } # no match at all!
    $year = Moving_Window($year);
    if (check_date($year,$month,$day))
    {
        return($year,$month,$day);
    }
    else { return(); } # not a valid date!
}

sub Decode_Date_US2
{
    die "Usage: (\$year,\$month,\$day) = Decode_Date_US2(\$date);\n"
      if (@_ != 1);

    my($buffer) = @_;
    my($year,$month,$day,$length);

    if ($buffer =~ /^[^A-Za-z0-9\xC0-\xD6\xD8-\xF6\xF8-\xFF]*  ([A-Za-z\xC0-\xD6\xD8-\xF6\xF8-\xFF]+)  [^A-Za-z0-9\xC0-\xD6\xD8-\xF6\xF8-\xFF]*  0*(\d+)  \D*$/x)
    {
        ($month,$buffer) = ($1,$2);
        $month = Decode_Month($month);
        unless ($month > 0)
        {
            return(); # can't decode month!
        }
        $length = length($buffer);
        if    ($length == 2)
        {
            $day  = substr($buffer,0,1);
            $year = substr($buffer,1,1);
        }
        elsif ($length == 3)
        {
            $day  = substr($buffer,0,1);
            $year = substr($buffer,1,2);
        }
        elsif ($length == 4)
        {
            $day  = substr($buffer,0,2);
            $year = substr($buffer,2,2);
        }
        elsif ($length == 5)
        {
            $day  = substr($buffer,0,1);
            $year = substr($buffer,1,4);
        }
        elsif ($length == 6)
        {
            $day  = substr($buffer,0,2);
            $year = substr($buffer,2,4);
        }
        else { return(); } # wrong number of digits!
    }
    elsif ($buffer =~ /^[^A-Za-z0-9\xC0-\xD6\xD8-\xF6\xF8-\xFF]*  ([A-Za-z\xC0-\xD6\xD8-\xF6\xF8-\xFF]+)  [^A-Za-z0-9\xC0-\xD6\xD8-\xF6\xF8-\xFF]*  (\d+)  \D+  (\d+)  \D*$/x)
    {
        ($month,$day,$year) = ($1,$2,$3);
        $month = Decode_Month($month);
        unless ($month > 0)
        {
            return(); # can't decode month!
        }
    }
    elsif ($buffer =~ /^\D*  0*(\d+)  \D*$/x)
    {
        $buffer = $1;
        $length = length($buffer);
        if    ($length == 3)
        {
            $month = substr($buffer,0,1);
            $day   = substr($buffer,1,1);
            $year  = substr($buffer,2,1);
        }
        elsif ($length == 4)
        {
            $month = substr($buffer,0,1);
            $day   = substr($buffer,1,1);
            $year  = substr($buffer,2,2);
        }
        elsif ($length == 5)
        {
            $month = substr($buffer,0,1);
            $day   = substr($buffer,1,2);
            $year  = substr($buffer,3,2);
        }
        elsif ($length == 6)
        {
            $month = substr($buffer,0,2);
            $day   = substr($buffer,2,2);
            $year  = substr($buffer,4,2);
        }
        elsif ($length == 7)
        {
            $month = substr($buffer,0,1);
            $day   = substr($buffer,1,2);
            $year  = substr($buffer,3,4);
        }
        elsif ($length == 8)
        {
            $month = substr($buffer,0,2);
            $day   = substr($buffer,2,2);
            $year  = substr($buffer,4,4);
        }
        else { return(); } # wrong number of digits!
    }
    elsif ($buffer =~ /^\D*  (\d+)  \D+  (\d+)  \D+  (\d+)  \D*$/x)
    {
        ($month,$day,$year) = ($1,$2,$3);
    }
    else { return(); } # no match at all!
    $year = Moving_Window($year);
    if (check_date($year,$month,$day))
    {
        return($year,$month,$day);
    }
    else { return(); } # not a valid date!
}

sub Parse_Date
{
    die "Usage: (\$year,\$month,\$day) = Parse_Date(\$date);\n"
      if (@_ != 1);

    my($date) = @_;
    my($year,$month,$day);
    unless ($date =~ /\b([JFMASOND][aepuco][nbrynlgptvc])\s+([0123]??\d)\b/)
    {
        return();
    }
    $month = $1;
    $day   = $2;
    unless ($date =~ /\b(19\d\d|20\d\d)\b/)
    {
        return();
    }
    $year  = $1;
    $month = Decode_Month($month);
    unless ($month > 0)
    {
        return();
    }
    unless (check_date($year,$month,$day))
    {
        return();
    }
    return($year,$month,$day);
}

1;

__END__

